# -*- coding: utf-8 -*-

import numpy as np
# import numpy.linalg as la
import scipy.linalg as sla
# import scipy.sparse as sps
# import warnings


def s_1l(thickness, ql, al0, bl0):
    """
    calculate the scattering matrix of 1 layer.

    Parameters
    ----------
    thickness   :   float
                    thickness of layer
    ql          :   ndarray
                    1d array, eigen propagation constant in z in layer
    al0         :   ndarray
    bl0         :   ndarray
                    interface matrix

    Returns
    -------
    s11, s12, s21, s22  :   ndarray
                            four elements of the scattering matrix

    """

    a = al0
    b = bl0
    alu = sla.lu_factor(a)

    fl = np.exp(1j * ql * thickness)

    t1 = a - fl[:, None] * b @ sla.lu_solve(alu, (fl[:, None] * b))
    t1lu = sla.lu_factor(t1)

    s21 = sla.lu_solve(t1lu, (fl[:, None] * (a - b @ sla.lu_solve(alu, b))))
    s22 = sla.lu_solve(t1lu, (fl[:, None] * b @ sla.lu_solve(alu, fl[:, None] * a) - b))

    s11 = s22
    s12 = s21

    return s11, s12, s21, s22


def s_1l_in(al0, bl0):
    """
    Calculate the scattering matrix of the 'input' region.
    The input region may not be vacuum. It has a thickness of infinity. The s-matrix is defined for the interface from this infinite material to (thickness 0) vacuum.

    Parameters
    ----------
    al0     :   ndarray
    bl0     :   ndarray

    Returns
    -------
    s11, s12, s21, s22  :   ndarray
    """

    a = al0
    b = bl0

    aTlu = sla.lu_factor(a.T)
    aTlu2 = (aTlu[0].copy(), aTlu[1].copy())
    a1 = aTlu2[0]
    a1[np.triu_indices(a1.shape[0])] *= 0.5
    alu = sla.lu_factor(a)
    ab = sla.lu_solve(alu, b)

    s11 = sla.lu_solve(aTlu, b.T).T
    s12 = 1./2. * (a - b @ ab)
    s21 = aTlu2
    s22 = - ab

    return s11, s12, s21, s22


def s_1l_out(al0, bl0):
    """
    Calculate the scattering matrix of the 'output' region.
    The output region may not be vacuum. It has a thickness of infinity. The s-matrix is defined for the interface from (thickness 0) vacuum to this infinite material.

    Parameters
    ----------
    al0     :   ndarray
    bl0     :   ndarray

    Returns
    -------
    s11, s12, s21, s22  :   ndarray
    """

    a = al0
    b = bl0

    alu = sla.lu_factor(a)
    aTlu = sla.lu_factor(a.T)
    aTlu2 = (aTlu[0].copy(), aTlu[1].copy())
    a1 = aTlu2[0]
    a1[np.triu_indices(a1.shape[0])] *= 0.5
    ab = sla.lu_solve(alu, b)

    s11 = - ab
    s12 = aTlu2
    s21 = 1./2. * (a - b @ ab)
    s22 = sla.lu_solve(aTlu, b.T).T

    return s11, s12, s21, s22





# -*- coding: utf-8 -*-

import numpy as np
from numpy import linalg as la
from typing import List, Union, Tuple, Optional
from warnings import warn


def gibbs_corr(ks: List[Union[float, Tuple[float, float], Tuple[float, float, float]]],
               m: Optional[float] = None,
               method: str = 'Gaussian',
               order: float = 1.,
               factor: float = 1.
               ) -> List[Union[float, Tuple[float, float], Tuple[float, float, float]]]:
    """
    Calculate the correction factor to mitigate Gibbs phenomenon.
    options are either Lanczos or Gaussian.

    Parameters
    ----------
    ks      :   the list of k points to calculate the correction factors
    m       :   defines the width of the averaging window for both Lanczos and gaussian. For standard Lanczos m is the largest k "+1".
    method  :   {'Gaussian', 'Lanczos'}, the method to use
    order   :   for Lanczos only; the power of the the sigma factor sinc term
    factor  :   if m not specified, can still multiply the calculated m by this factor.

    Returns
    -------
    s       :   the correction coefficients for gibbs

    """

    if len(ks) == 1 and (m is None):
        warn('Only one k point is given, with no m specified, can not calculate the correction factor.', UserWarning)
        s = np.array([1])
    else:
        if m == 0.:
            raise Exception('m can not be zero')

        ksa = np.array(ks)

        # calculate the norms of the k's
        if ksa.ndim == 1:
            kn = np.abs(ksa)
        else:
            ax = tuple(i for i in np.arange(ksa.ndim)[1:])  # e.g. for 3d k, ksa is 4d, and this gives (1, 2, 3)
            kn = la.norm(ksa, axis=ax)  # k norm

        # calculate m if not given
        if m is None:
            knp = np.partition(kn, -2)
            m = kn.max() + (knp[-1] - knp[-2])
            if method == 'Gaussian':
                m *= 0.7

        m *= factor
        ma = kn / m

        if method == 'Lanczos':
            s = (np.sinc(ma)) ** order
        else:
            s = np.exp(-ma**2)

    return s.tolist()





# -*- coding: utf-8 -*-

from typing import Tuple, Union
import numpy as np
# import numpy.linalg as la
# from warnings import warn


class Mtr:
    # material need to have a name such that user can access it by its name
    def __init__(self,
                 epsi: Union[Union[float, complex], Tuple[Union[float, complex], Union[float, complex], Union[float, complex]], np.ndarray],
                 mu: Union[Union[float, complex], Tuple[Union[float, complex], Union[float, complex], Union[float, complex]], np.ndarray],
                 name=None):
        """
        Material.
        """
        self._epsi = None
        self._mu = None
        self._epsi_inv = None
        self._mu_inv = None

        self.epsi = epsi
        self.mu = mu

        self.name = name

    @property
    def epsi(self) -> np.ndarray:
        return self._epsi

    @epsi.setter
    def epsi(self, val):
        if (type(val) is float) or (type(val) is int) or (type(val) is complex) or (type(val) is np.float64) or (type(val) is np.complex128):
            val = np.eye(3, dtype=complex) * val
        elif np.array(val).ndim == 1 and np.array(val).size == 3:
            val = np.diag(val) + 0j
        elif (type(val) is not np.ndarray) or val.shape != (3, 3):
            raise ValueError('data format for epsilon is incorrect.')
        self._epsi = val + 0j
        # explicit computation of inverse.
        v = val
        adbc = v[0, 0] * v[1, 1] - v[0, 1] * v[1, 0]
        if adbc == 0 or v[2, 2] == 0:
            raise Exception('Singular permittivity tensor.')
        ei = np.array([[v[1, 1] / adbc, -v[0, 1] / adbc, 0],
                       [-v[1, 0] / adbc, v[0, 0] / adbc, 0],
                       [0, 0, 1 / v[2, 2]]], dtype=complex)
        self._epsi_inv = ei

    @property
    def mu(self) -> np.ndarray:
        return self._mu

    @mu.setter
    def mu(self, val):
        if (type(val) is float) or (type(val) is int) or (type(val) is complex) or (type(val) is np.float64) or (type(val) is np.complex128):
            val = np.eye(3, dtype=complex) * val
        elif np.array(val).ndim == 1 and np.array(val).size == 3:
            val = np.diag(val) + 0j
        elif (type(val) is not np.ndarray) or val.shape != (3, 3):
            raise ValueError('data format for mu is incorrect.')
        self._mu = val + 0j
        # explicit computation of inverse.
        v = val
        adbc = v[0, 0] * v[1, 1] - v[0, 1] * v[1, 0]
        if adbc == 0 or v[2, 2] == 0:
            raise Exception('Singular permittivity tensor.')
        mi = np.array([[v[1, 1] / adbc, -v[0, 1] / adbc, 0],
                       [-v[1, 0] / adbc, v[0, 0] / adbc, 0],
                       [0, 0, 1 / v[2, 2]]], dtype=complex)
        self._mu_inv = mi

    @property
    def epsi_inv(self) -> np.ndarray:
        return self._epsi_inv

    @property
    def mu_inv(self) -> np.ndarray:
        return self._mu_inv




# -*- coding: utf-8 -*-

import numpy as np
# from numpy import linalg as la


def g_pts(num_g, b1, b2):
    """
    given number of lattice points, and two lattice vectors, give the list of all lattice points inside a circle. the total number of lattice points returned is roughly the given number of lattice points.

    This function is indifferent whether b1 and b2 are primitive lattice or reciprocal lattice.

    Parameters
    ----------
    num_g       :   int
                    target total number of lattice points
    b1, b2      :   tuple[float, float]
                    lattice vectors. For rectangular lattice, the input b vectors should be orthogonal

    Returns
    -------
    k_pts       :   list[tuple[float, float]]
                    the k points
    idx         :   list[tuple[int, int]]
                    the indices of the k points, i.e. (m, n) as in m*b1 + n*b2 is the corresponding k points.

    """

    b1 = np.array(b1)
    b2 = np.array(b2)

    bz_are = np.abs(np.cross(b1, b2))  # Brillouin zone area

    k_radi = np.sqrt(num_g * bz_are / np.pi)  # k points within the vertices are to be included

    k_pts = []
    idx = []

    if b1 @ b2 == 0:  # rectangular lattice
        l1 = np.linalg.norm(b1)
        M = np.ceil(k_radi / l1).astype(int)
        l2 = np.linalg.norm(b2)
        N = np.ceil(k_radi / l2).astype(int)
        # m and n array for quarter parallelogram with edges included
        m_a = np.arange(0, M+1)
        n_a = np.arange(0, N+1)
        m_grid, n_grid = np.meshgrid(m_a, n_a)  # quarter parallelogram, with edges included
        # k points coordinates in the quarter parallelogram
        kkx = m_grid * b1[0] + n_grid * b2[0]
        kky = m_grid * b1[1] + n_grid * b2[1]
        k_dis = np.sqrt(kkx ** 2 + kky ** 2)  # 2d array containing distances
        # which k point in the half parallelogram is inside the circle
        id_n, id_m = np.where(k_dis <= k_radi)  # each is a 1d array. containing the m and n index of the points inside. Note numpy default indexing is row for x column for y.
        for i1, i2 in zip(id_n, id_m):
            if m_a[i2] == 0 and n_a[i1] == 0:
                k_pts.append((0., 0.))
                idx.append((0, 0))
            elif m_a[i2] == 0 or n_a[i1] == 0:  # add these points on the two arms of the quarter parallelogram and their opposite points.
                k = (kkx[i1, i2], kky[i1, i2])
                k_i = (-kkx[i1, i2], -kky[i1, i2])  # opposite point
                k_pts.append(k)
                k_pts.append(k_i)
                idx.append((m_a[i2], n_a[i1]))
                idx.append((-m_a[i2], -n_a[i1]))
            else:  # for points not on the two arms, add them and their 3 symmetry partners
                k = (kkx[i1, i2], kky[i1, i2])
                k1 = (-kkx[i1, i2], kky[i1, i2])
                k2 = (kkx[i1, i2], -kky[i1, i2])
                k3 = (-kkx[i1, i2], -kky[i1, i2])
                k_pts += [k, k1, k2, k3]
                idx += [(m_a[i2], n_a[i1]), (-m_a[i2], n_a[i1]), (m_a[i2], -n_a[i1]), (-m_a[i2], -n_a[i1])]
    else:
        l1 = bz_are / np.linalg.norm(b2)
        M = np.ceil(k_radi / l1).astype(int)
        l2 = bz_are / np.linalg.norm(b1)
        N = np.ceil(k_radi / l2).astype(int)
        # m and n array for half parallelogram with edges included
        m_a = np.arange(-M, M+1)
        n_a = np.arange(0, N+1)
        m_grid, n_grid = np.meshgrid(m_a, n_a)  # half parallelogram, with edges included
        # k points coordinates in the half parallelogram
        kkx = m_grid * b1[0] + n_grid * b2[0]
        kky = m_grid * b1[1] + n_grid * b2[1]
        k_dis = np.sqrt(kkx ** 2 + kky ** 2)  # 2d array containing distances
        # which k point in the half parallelogram is inside the circle
        id_n, id_m = np.where(k_dis <= k_radi)  # each is a 1d array. containing the index of m and n of the points inside
        k_pts.append((0., 0.))
        idx.append((0, 0))
        for i1, i2 in zip(id_n, id_m):
            if not (m_a[i2] <= 0 and n_a[i1] == 0):  # excluding this half "arm" to avoid duplication. n index = 0 and m index > 0 is already included.
                k = (kkx[i1, i2], kky[i1, i2])
                k_i = (-kkx[i1, i2], -kky[i1, i2])  # opposite point
                k_pts.append(k)
                k_pts.append(k_i)
                idx.append((m_a[i2], n_a[i1]))
                idx.append((-m_a[i2], -n_a[i1]))

    return k_pts, idx




# -*- coding: utf-8 -*-

"""
Calculate the Fourier transform of 2d polygon.

Reference:
K. McInturff and P.S. Simon, "The Fourier transform of linearly varying functions with polygonal support", IEEE Trans. Ann. Prop.  39, 1441 (1991)
"""

import numpy as np
import numpy.linalg as la
from scipy.special import jn
from inkstone.ft.poly_area import poly_area


def ft_2d_poly_1(vertices, ks):
    """
    Calculate the fourier transform of a function with value 1 inside a polygon and 0 outside. Assuming none of the required k is at (0, 0)

    Parameters
    ----------
    vertices    :   list[tuple[float, float]] or ndarray
                    vertices of the polygon, in counterclockwise sequence.
                    Can give a list of the coordinates, each vertex is (x, y) tuple. Or, give a in nx2 shaped ndarray
    ks          :   list[tuple[float, float]] or ndarray
                    a list of k vectors. The k vectors can't be (0, 0)
                    can give a list of k's, each k is a (kx, ky) tuple. Or, give a mx2 shaped ndarray

    Returns
    -------
    s           :   list[complex]
                    1d array, value of Fourier coefficient at the corresponding ks points.

    """
    # convert to array
    ksa = np.array(ks)  # each row is a [kx, ky]

    ks_nm = la.norm(ksa, axis=-1)
    coef = -1j / ks_nm ** 2

    # convert tuples to numpy arrays
    verti = np.array(vertices)

    # generate a shifted array of vertices of [r2, r3, ..., rn, r1], if originally [r1, r2, ..., rn]
    verti_roll = np.roll(verti, -1, axis=0)

    rnc = (verti + verti_roll) / 2.

    ln = verti_roll - verti

    cross = np.cross(ln[:, None, :], ksa[None, :, :])
    # Say ln.shape is mx2, ksa is nx2, this gives mxn shape

    term1 = cross * (np.exp(-1j * rnc @ ksa.T)) * jn(0, ln @ ksa.T / 2.)

    s = coef * np.sum(term1, axis=0)

    return s.tolist()


def ft_2d_poly(vertices, ks):
    """
    Calculate the Fourier transform of a function with value 1 inside a polygon shape and 0 outside.

    Parameters
    ----------
    vertices    :   list[tuple[float, float]]
                    vertices of the polygon, in counterclockwise sequence.
    ks          :   list[tuple[float, float]]
                    a list of k vectors.

    Returns
    -------
    s           :   list[complex]
                    1d array, value of Fourier coefficient at the corresponding ks points.

    Notes
    -----
    For k == (0, 0) and k != (0, 0), call different subroutine.
    """
    ksa = np.array(ks)  # convert to array
    ks_nm = la.norm(ksa, axis=-1)  # calculate the norm of each k vector
    idx_0 = np.where(ks_nm == 0)[0]  # index to where k is (0, 0)
    idx_i = np.where(ks_nm != 0)[0]  # index to where k is not (0, 0)
    ksa1 = np.delete(ksa, idx_0, axis=0)  # new ks array that doesn't contain (0, 0)

    s1 = np.array(ft_2d_poly_1(vertices, ksa1))
    a = poly_area(vertices)

    s = 1j * np.zeros(ks_nm.size)
    s[idx_i] = s1
    s[idx_0] = a

    return s.tolist()




# -*- coding: utf-8 -*-

import numpy as np
from numpy import linalg as la
# import scipy.sparse as sps
import scipy.fft as fft
import scipy.linalg as sla
from warnings import warn
from collections import OrderedDict
from typing import Optional, Set, Union, Tuple, Dict, List
import time

from inkstone.ft.ft_2d_cnst import ft_2d_cnst
from inkstone.im import im
from inkstone.sm import s_1l, s_1l_in, s_1l_out
from inkstone.params import Params
from inkstone.bx import Bx
from inkstone.mtr import Mtr
from inkstone.shps import Rect, Para, Elli, Disk, Poly, OneD
from inkstone.helpers.pt_in_poly import pt_in_poly


class Layer:
    # todo: for uniform layer should use sparse matrices, faster less memory

    def __init__(self, name, thickness, material_bg, materials, params, **kwargs):
        """
        A layer.

        One can add patterns to the layer. The layer calculates the fourier series and solve for the scattering matrix of the layer.

        Parameters
        ----------
        name        :   str
                        the name of the layer
        thickness   :   float
                        thickness of the layer. for input (reflected side) region or output (transmitted side) region, the thickness has to be zero.
        material_bg :   str
        materials   :   dict[str, Mtr]
                        all of the materials used in the entire structure.
        params      :   Params
        kwargs      :
                        keyword arguments to pass on to SetLayer

        """

        self.pr: Params = params

        self.if_mod: bool = True  # if this layer is modified
        self.if_t_change: bool = True  # if thickness changed
        self.need_recalc_al_bl = True  # if the al and bl of this layer needs recalculation

        self.name: str = name

        self.thickness: float = thickness

        self.material_bg: str = material_bg  # background material
        self.materials: Dict[str, Mtr] = materials  # all of the materials used in the structure

        self.materials_used: Set[str] = set()  # all the materials used by this layer
        self.materials_used.add(material_bg)
        mb = self.materials[material_bg]
        if mb.epsi[0, 1] != 0 or mb.epsi[1, 0] != 0. or mb.epsi[0, 1] != 0. or mb.mu[0, 1] != 0. or mb.mu[1, 0] != 0.:
            self.materials_ode: bool = True  # if any of the materials used have off-diagonal components
        else:
            self.materials_ode = False

        self.patterns: OrderedDict[str, Bx] = OrderedDict()  # all patterns of this layer

        # for 2d and 3d
        self.epsi_fs: Optional[np.ndarray] = None  # Fourier series components of epsilon in the layer, shape (2mmax+1, 2nmax+1, 3, 3) complex, not ifftshifted
        self.epsi_inv_fs: Optional[np.ndarray] = None
        #
        self.mu_fs: Optional[np.ndarray] = None  # shape (2mmax+1, 2nmax+1, 3, 3) complex, not ifftshifted
        self.mu_inv_fs: Optional[np.ndarray] = None
        #
        # the components that are actually used in convolution matrices
        self.epsi_fs_used: Optional[List[np.ndarray]] = None  # each element is (3, 3) shape
        self.epsi_inv_fs_used: Optional[List[np.ndarray]] = None
        self.mu_fs_used: Optional[List[np.ndarray]] = None
        self.mu_inv_fs_used: Optional[List[np.ndarray]] = None

        # for 3d epsilon and mu convolution matrices
        self.epxxcm, self.epxycm, self.epyxcm, self.epyycm, self.epzzcm, \
        self.eixxcm, self.eixycm, self.eiyxcm, self.eiyycm, self.eizzcm, \
        self.muxxcm, self.muxycm, self.muyxcm, self.muyycm, self.muzzcm, \
        self.mixxcm, self.mixycm, self.miyxcm, self.miyycm, self.mizzcm, \
            = [None for i in range(20)]  # shape （num_g, num_g）

        # --- variables require solving of layer ---

        self.ql: Optional[np.ndarray] = None  # 1d array complex
        self.phil: Optional[np.ndarray] = None  # shape (2num_g, 2num_g)
        self._phil_is_idt = False  # whether phil is identity matrix
        self.psil: Optional[np.ndarray] = None

        self.im: Optional[Tuple[np.ndarray, np.ndarray]] = None  # interface matrix
        self.sm: Optional[Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]] = None  # scattering matrix
        self.csm: Optional[Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]] = None  # cumulative scattering matrix
        self.csmr: Optional[Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]] = None  # cumulative scattering matrix reversed

        self.al_bl: Optional[Tuple[np.ndarray, np.ndarray]] = None  # the field coefficients (al, bl).

        self.in_mid_out: str = 'mid'  # {'in', 'mid', 'out'}, if this layer is the incident, output, or a middle layer

        if material_bg == 'vacuum':
            self.is_vac = True
        else:
            self.is_vac = False

        self._rad_cha: Optional[List[int]] = None

        self.set_layer(**kwargs)

    @property
    def rad_cha(self):
        """List of indices in `.params.Params.idx_g` that corresponds to radiation channels in this layer"""
        # For 1 channel with two polarizations this would be [0, num_g]
        # if self.is_vac:
        #     return self.pr.rad_cha_0
        # else:
        return self._rad_cha

    def set_layer(self,
                  thickness: float = None,
                  material_bg: str = None
                  ):
        """
        Set and reset parameters of the layer.
        """
        if (thickness is not None) and (self.thickness != thickness):
            self.thickness = thickness
            self.if_t_change = True
        if (material_bg is not None) and (self.material_bg != material_bg):
            self.material_bg = material_bg
            self.if_mod = True
        if material_bg == 'vacuum' and not self.patterns:
            self.is_vac = True

    def add_box(self,
                mtr: str,
                shp: str,
                box_name: str = None,
                **kwargs):
        """
        Add a box to the layer.

        Parameters
        ----------
        mtr         :
        shp         :
        box_name    :
        kwargs      :
                        other parameters to pass on to `~.bx.Bx`
        """
        self.materials_used.add(mtr)
        mt = self.materials[mtr]
        if mt.epsi[0, 1] != 0 or mt.epsi[1, 0] != 0. or mt.epsi[0, 1] != 0. or mt.mu[0, 1] != 0. or mt.mu[1, 0] != 0.:
            self.materials_ode: bool = True  # if any of the materials used have off-diagonal components

        if box_name is None:
            box_name = 'box{:d}'.format(len(self.patterns))
        if box_name in self.patterns:
            warn('A box with this name already exists in this layer. The existing box will be overridden.', UserWarning)
            # todo: remove the material used in the overridden box. be careful not to remove materials still in use by other boxes

        bx = Bx(mt, shp, name=box_name, **kwargs)
        bx.ks = self.pr.ks_ep_mu
        self.patterns[box_name] = bx
        self._find_bx_outside()
        self.if_mod = True
        if mtr != 'vacuum':
            self.is_vac = False

    def set_box(self,
                box_name: str,
                **kwargs):
        """
        set the parameters of an existing box

        Parameters
        ----------
        box_name        :
        kwargs          :
                            shape parameters to reset for box, to be passed on to `.bx.Bx.set_shape`

        Returns
        -------

        """
        if box_name in self.patterns:
            box: Bx = self.patterns[box_name]
            box.set_shape(**kwargs)
            self._find_bx_outside()
            self.if_mod = True
            # todo: allow change the type of material of the box
        else:
            warn('Did not find the name of the box you want to change.', UserWarning)

    def _find_bx_outside(self):
        """ find out which box is in which """

        bxs = list(self.patterns.values())

        bx_areas = np.array([a.shp.area for a in bxs])
        idx = np.argsort(bx_areas)
        bx_names = np.array([a.name for a in bxs])
        bx_name_sorted = bx_names[idx]

        # a fictional bx with background material of this layer.
        bxf = Bx(self.materials[self.material_bg], 'polygon', name='the cell', vertices=[(0, 0), (1, 0), (0, 1)])

        if self.pr.is_1d_latt:
            for ii, name1 in enumerate(bx_name_sorted):
                bx1 = self.patterns[name1]

                # default outside is the fictional bx
                bx1.outside = bxf

                pt1: float = bx1.shp.center

                for name2 in bx_name_sorted[ii + 1:]:
                    bx2 = self.patterns[name2]

                    w2 = bx2.shp.width
                    c2 = bx2.shp.center
                    if c2 - w2 / 2. < pt1 < c2 + w2 / 2.:
                        bx1.outside = bx2
                        break
        else:  # 3D
            for ii, name1 in enumerate(bx_name_sorted):
                bx1 = self.patterns[name1]
                # default outside is the fictional bx
                bx1.outside = bxf

                # find a pt1 that's inside bx1
                if bx1.shp.shape == 'rectangle' or bx1.shp.shape == 'disk' or bx1.shp.shape == 'ellipse' or bx1.shp.shape== 'parallelogram':
                    pt1: Tuple[float, float] = bx1.shp.center
                elif bx1.shp.shape == 'polygon':
                    vertices = bx1.shp.vertices
                    v1, v2, v3 = np.array([vertices[i] for i in range(3)])
                    vm = (v3 + v1) / 2.
                    r = vm - v2
                    while True:
                        r /= -2.
                        v_new = r + v2
                        if pt_in_poly(vertices, v_new):
                            break
                    pt1: Tuple[float, float] = (v_new[0], v_new[1])
                else:
                    raise Exception("Shape of pattern not recognized.")

                for name2 in bx_name_sorted[ii + 1:]:
                    bx2 = self.patterns[name2]

                    # see if pt1 is inside bx2
                    if bx2.shp.shape == 'rectangle':
                        r = np.array(pt1) - np.array(bx2.shp.center)
                        a = bx2.shp.angle * np.pi / 180.
                        r1 = np.array([[np.cos(a), np.sin(a)], [-np.sin(a), np.cos(a)]]) @ r
                        if (r1[0] < bx2.shp.side_lengths[0] / 2.) and (r1[1] < bx2.shp.side_lengths[1] / 2.):
                            bx1.outside = bx2
                            break
                    elif bx2.shp.shape == 'parallelogram':
                        shp: Para = bx2.shp
                        r = np.array(pt1) - np.array(shp.center)
                        a = shp.angle * np.pi / 180.
                        rot = np.array([[np.cos(a), np.sin(a)], [-np.sin(a), np.cos(a)]])
                        m = np.tan(np.pi / 2 - shp.shear_angle * np.pi / 180)
                        sheer = np.array([[1, -m],
                                          [0, 1]])
                        r1 = sheer @ rot @ r
                        if (np.abs(r1[0]) < shp.side_lengths[0] / 2.) and (np.abs(r1[1]) < shp.side_lengths[1] * np.sin(shp.shear_angle * np.pi / 180) / 2.):
                            bx1.outside = bx2
                            break
                    elif bx2.shp.shape == 'disk':
                        r = np.array(pt1) - np.array(bx2.shp.center)
                        d = np.linalg.norm(r)
                        if d < bx2.shp.radius:
                            bx1.outside = bx2
                            break
                    elif bx2.shp.shape == 'ellipse':
                        r = np.array(pt1) - np.array(bx2.shp.center)
                        a = bx2.shp.angle * np.pi / 180.
                        r1 = np.array([[np.cos(a), np.sin(a)], [-np.sin(a), np.cos(a)]]) @ r
                        if ((r1[0] ** 2 / bx2.shp.half_widths[0] ** 2 + r1[1] ** 2 / bx2.shp.half_widths[1] ** 2) < 1):
                            bx1.outside = bx2
                            break
                    elif bx2.shp.shape == 'polygon':
                        vs = bx2.shp.vertices
                        if pt_in_poly(vs, pt1):
                            bx1.outside = bx2
                            break
                    else:
                        raise Exception("Shape of pattern not recognized.")

    def _calc_ep_mu_fs_3d(self):
        """ calculate the fourier coefficients of this layer """
        # todo: implement add material permittivity and permeability complex tensors
        # todo: for uniform layer, is it simpler?

        t1 = time.process_time()

        # calculate the Fourier components of the background material.
        mtr = self.materials[self.material_bg]
        epsi_bg, epsi_bg_inv, mu_bg, mu_bg_inv = [mtr.epsi, mtr.epsi_inv, mtr.mu, mtr.mu_inv]  # complex 3x3 tensors
        d = np.array(ft_2d_cnst(self.pr.ks_ep_mu), dtype=complex)

        ep, ei, mu, mi = [t[None, :, :] * d[:, None, None] for t in [epsi_bg, epsi_bg_inv, mu_bg, mu_bg_inv]]   # each is complex ((2mmax+1)x(2nmax+1), 3, 3) shape

        for bx in self.patterns.values():
            eb, eib, mb, mib = [np.array(f, dtype=complex) / self.pr.uc_area for f in bx.ft(self.pr.ks_ep_mu)]
            ep += eb
            ei += eib
            mu += mb
            mi += mib

        d1 = 2 * self.pr.mmax + 1
        d2 = 2 * self.pr.nmax + 1
        epa, eia, mua, mia = [a.reshape(d2, d1, 3, 3) for a in [ep, ei, mu, mi]]  # complex ((2nmax+1), (2mmax+1), , 3, 3) shape
        self.epsi_fs = epa
        self.epsi_inv_fs = eia
        self.mu_fs = mua
        self.mu_inv_fs = mia

        if self.pr.show_calc_time:
            print("{:.6f}   _calc_ep_mu_fs_3d".format(time.process_time() - t1) + ", layer "+self.name)

    def _cons_ep_mu_cm_3d(self):
        """ Construct epsilon and mu convolution matrices """
        t1 = time.process_time()

        if self.patterns:
            idx = self.pr.idx_conv_mtx
            ems = [fft.ifftshift(self.epsi_fs.swapaxes(0, 1), axes=(0, 1)),
                   fft.ifftshift(self.epsi_inv_fs.swapaxes(0, 1), axes=(0, 1)),
                   fft.ifftshift(self.mu_fs.swapaxes(0, 1), axes=(0, 1)),
                   fft.ifftshift(self.mu_inv_fs.swapaxes(0, 1), axes=(0, 1))]
            self.epxxcm, self.epxycm, self.epyxcm, self.epyycm, self.epzzcm, \
            self.eixxcm, self.eixycm, self.eiyxcm, self.eiyycm, self.eizzcm, \
            self.muxxcm, self.muxycm, self.muyxcm, self.muyycm, self.muzzcm, \
            self.mixxcm, self.mixycm, self.miyxcm, self.miyycm, self.mizzcm, \
                = [em[idx[:, :, 0], idx[:, :, 1], i, j]
                   for em in ems
                   for i, j in [(0, 0), (0, 1), (1, 0), (1, 1), (2, 2)]]
        else:
            # uniform layer
            mtr = self.materials[self.material_bg]
            epsi_bg, epsi_bg_inv, mu_bg, mu_bg_inv = [mtr.epsi, mtr.epsi_inv, mtr.mu, mtr.mu_inv]  # complex 3x3 tensors
            # d = np.eye(self.pr.num_g)
            self.epxxcm, self.epxycm, self.epyxcm, self.epyycm, self.epzzcm, \
            self.eixxcm, self.eixycm, self.eiyxcm, self.eiyycm, self.eizzcm, \
            self.muxxcm, self.muxycm, self.muyxcm, self.muyycm, self.muzzcm, \
            self.mixxcm, self.mixycm, self.miyxcm, self.miyycm, self.mizzcm, \
                = [np.diag(np.full(self.pr.num_g, em[i, j], dtype=complex)) for em in [epsi_bg, epsi_bg_inv, mu_bg, mu_bg_inv] for i, j in [(0, 0), (0, 1), (1, 0), (1, 1), (2, 2)]]

        if self.pr.show_calc_time:
            print("{:.6f}   _cons_ep_mu_cm_3d".format(time.process_time() - t1) + ", layer "+self.name)

    def reconstruct(self,
                    n1: int = None,
                    n2: int = None) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """

        Parameters
        ----------
        n1      :   number of points in the first lattice vector direction
        n2      :   number of points in the second lattice vector direction

        Returns
        -------
        xx          :   ndarray
                        x coordinates in a unit cell (2d ndarray), shape (n2, n1)
        yy          :   ndarray
                        y coordinates in a unit cell (2d ndarray), shape (n2, n1)
        epsi_recons :   ndarray
                        reconstructed epsi in a unit cell, shape (n2, n1, 3, 3)
        mu_recons   :   ndarray
                        reconstructed mu in a unit cell, shape (n2, n1, 3, 3)

        Notes
        -----
        Unit cell can be slanted. Generally xx and yy defines the grid points. can use matplotlib.pyplot to show the epsilon and mu profile, such as

            plt.pcolormesh(xx, yy, epsi_recons[:, :, 0, 0]

        to show the xx component of epsilon in a unit cell.
        """
        if n1 is None:
            n1 = 101
        if n2 is None:
            n2 = 101
        self.epsi_fs_used, self.epsi_inv_fs_used, self.mu_fs_used, self.mu_inv_fs_used = \
            [[em[j, i, :, :] for i, j in self.pr.idx_g_ep_mu_used]
             for em in [fft.ifftshift(self.epsi_fs, axes=(0, 1)), fft.ifftshift(self.epsi_inv_fs, axes=(0, 1)), fft.ifftshift(self.mu_fs, axes=(0, 1)), fft.ifftshift(self.mu_inv_fs, axes=(0, 1))]]

        eu = self.epsi_fs_used
        ea = np.array(eu, dtype=complex)
        mu = self.mu_fs_used
        ma = np.array(mu, dtype=complex)
        idx = self.pr.idx_g_ep_mu_used
        idxa = np.array(idx)

        # spatial coordinates in a unit cell of any shape
        x1 = np.linspace(-0.5, 0.5, n1)
        x2 = np.linspace(-0.5, 0.5, n2)
        xx1, xx2 = np.meshgrid(x1, x2)
        a1, a2 = [np.array(lv) for lv in self.pr.latt_vec]
        xx = xx1 * a1[0] + xx2 * a2[0]
        yy = xx1 * a1[1] + xx2 * a2[1]

        exp_term = np.exp(1j * 2 * np.pi *
                          (idxa[:, 0][:, None, None] * xx1[None, :, :] +
                           idxa[:, 1][:, None, None] * xx2[None, :, :]))
        epsi_recons = np.sum(ea[:, None, None, :, :] * exp_term[:, :, :, None, None], axis=0)
        mu_recons = np.sum(ma[:, None, None, :, :] * exp_term[:, :, :, None, None], axis=0)
        return xx, yy, epsi_recons, mu_recons

    def _calc_PQ_3d(self):
        """Calculate the P and Q matrix"""

        t1 = time.process_time()

        o = self.pr.omega
        epxx, epxy, epyx, epyy, epzz, \
        eixx, eixy, eiyx, eiyy, eizz, \
        muxx, muxy, muyx, muyy, muzz, \
        mixx, mixy, miyx, miyy, mizz, \
            = [self.epxxcm, self.epxycm, self.epyxcm, self.epyycm, self.epzzcm,
               self.eixxcm, self.eixycm, self.eiyxcm, self.eiyycm, self.eizzcm,
               self.muxxcm, self.muxycm, self.muyxcm, self.muyycm, self.muzzcm,
               self.mixxcm, self.mixycm, self.miyxcm, self.miyycm, self.mizzcm]
        Kx = self.pr.Kx
        Ky = self.pr.Ky

        P11 = o * muyx + 1. / o * Kx[:, None] * eizz * Ky
        P12 = o * muyy - 1. / o * Kx[:, None] * eizz * Kx
        P21 = -o * muxx + 1. / o * Ky[:, None] * eizz * Ky
        P22 = -o * muxy - 1. / o * Ky[:, None] * eizz * Kx

        Q11 = o * epyx + 1. / o * Kx[:, None] * mizz * Ky
        Q12 = o * epyy - 1. / o * Kx[:, None] * mizz * Kx
        Q21 = -o * epxx + 1. / o * Ky[:, None] * mizz * Ky
        Q22 = -o * epxy - 1. / o * Ky[:, None] * mizz * Kx

        P = np.block([[P11, P12],
                      [P21, P22]])
        Q = np.block([[Q11, Q12],
                      [Q21, Q22]])

        self.P = P
        self.Q = Q

        if self.pr.show_calc_time:
            print("{:.6f}   _calc_PQ_3d".format(time.process_time() - t1) + ", layer "+self.name)

    def _calc_eig_3d(self):
        """ Calculate the eigen modes in the layer """

        t1 = time.process_time()

        ql2, self.phil = la.eig(- self.P @ self.Q)
        self._rad_cha = np.where(ql2.real > 0)[0].tolist()  # todo: even for radiation channel, if omega.imag larger than omega.real, q02.real is negative
        ql = np.sqrt(ql2 + 0j)
        # ql[(ql2.real < 0) * (ql.imag < 0)] *= -1
        ql[ql.imag < 0] *= -1
        self.ql = ql

        ql_inv = 1. / self.ql
        self.psil = -1j * self.Q @ self.phil * ql_inv
        # self.psil = 1j * sla.solve(self.P, self.phil) @ np.diag(self.ql)

        if self.pr.show_calc_time:
            print("{:.6f}   _calc_eig_3d".format(time.process_time() - t1) + ", layer "+self.name)

    def _calc_eig_3d_uniform(self):
        """ Efficient calculation of eigen for uniform layer """

        t1 = time.process_time()

        if self.is_vac:
            self.ql = self.pr.q0
            self.phil = self.pr.phi0
            self.psil = self.pr.psi0
            self._rad_cha = self.pr.rad_cha_0 + [i+self.pr.num_g for i in self.pr.rad_cha_0]
        else:  # for numG 100 this takes about 2ms
            kxa, kya = [a.ravel() for a in np.hsplit(np.array(self.pr.ks), 2)]
            o = self.pr.omega
            mtr = self.materials[self.material_bg]
            mxx, mxy, myx, myy, mzz, exx, exy, eyx, eyy, ezz = [a[i, j] for a in [mtr.epsi, mtr.mu] for i, j in [(0, 0), (0, 1), (1, 0), (1, 1), (2, 2)]]
            p = np.array([[o * myx + 1./o / ezz * kxa * kya, o * myy - 1./o / ezz * kxa**2],
                          [-o * mxx + 1./o / ezz * kya**2, -o * mxy - 1./o / ezz * kxa * kya]], dtype=complex)  # (2, 2, num_g) shape
            q = np.array([[o * eyx + 1./o / mzz * kxa * kya, o * eyy - 1./o / mzz * kxa**2],
                          [-o * exx + 1./o / mzz * kya**2, -o * exy - 1./o / mzz * kxa * kya]], dtype=complex)  # (2, 2, num_g) shape
            p = np.rollaxis(p, -1)  # shape (num_g, 2, 2)
            q = np.rollaxis(q, -1)  # shape (num_g, 2, 2)
            pq = p @ q  # shape (num_g, 2, 2)

            if (np.abs(mxy) + np.abs(myx) + np.abs(exy) + np.abs(eyx)) == 0. and (eyy/ezz == myy/mzz) and (exx/ezz == mxx/mzz):
                # direct construction of eigen, no solving
                self._phil_is_idt = True
                phil = np.eye(2*self.pr.num_g, dtype=complex)
                v = np.eye(2, dtype=complex)[None, :, :]
                w2 = - (pq[:, range(2), range(2)]).T.ravel()  # shape (num_g, 2) then ravel out
                self._rad_cha = np.where(w2.real > 0)[0].tolist()  # todo: even for radiation channel, if omega.imag larger than omega.real, q02.real is negative
                w = np.sqrt(w2 + 0j)

            else:
                w2, v = la.eig(-pq)  # w2 shape (num_g, 2), v shape (num_g, 2, 2)
                w2 = w2.T.ravel()  # shape (num_g, 2) then ravel out
                self._rad_cha = np.where(w2.real > 0)[0].tolist()  # todo: even for radiation channel, if omega.imag larger than omega.real, q02.real is negative
                w = np.sqrt(w2 + 0j)

                phil11, phil12, phil21, phil22 = [v[:, i, j] for i, j in [(0, 0), (0, 1), (1, 0), (1, 1)]]
                phil = np.zeros((2*self.pr.num_g, 2*self.pr.num_g), dtype=complex)
                r1 = range(self.pr.num_g)
                r2 = range(self.pr.num_g, 2 * self.pr.num_g)
                phil[r1, r1] = phil11
                phil[r1, r2] = phil12
                phil[r2, r1] = phil21
                phil[r2, r2] = phil22

            if self.in_mid_out == 'mid':
                w[w.imag < 0] *= -1
            else:
                if self.pr.omega.imag < 0:
                    if self.pr.ccnif == "physical":
                        w[(w2.real < 0) * (w.imag < 0)] *= -1
                    elif self.pr.ccnif == "ac":
                        w[(w2.real < 0) * (w.imag > 0)] *= -1
                    else:
                        warn("ccnif not recognized. Default to 'physical'.")
                        w[(w2.real < 0) * (w.imag < 0)] *= -1
                elif self.pr.omega.imag > 0:
                    if self.pr.ccpif == "ac":
                        w[(w2.real < 0) * (w.imag < 0)] *= -1
                    elif self.pr.ccpif == "physical":
                        w[(w2.real < 0) * (w.imag > 0)] *= -1
                    else:
                        warn("ccpif not recognized. Default to 'ac'.")
                        w[(w2.real < 0) * (w.imag < 0)] *= -1
                else:
                    w[w.imag < 0] *= -1
            ql = w  # 1d array length 2num_g
            vh = -1j * p @ v / w[:, None]

            psil11, psil12, psil21, psil22 = [np.diag(vh[:, i, j]) for i, j in [(0, 0), (0, 1), (1, 0), (1, 1)]]
            psil = np.block([[psil11, psil12],
                             [psil21, psil22]])

            self.ql = ql
            self.phil = phil
            self.psil = psil

        if self.pr.show_calc_time:
            print("{:.6f}   _calc_eig_3d_uniform".format(time.process_time() - t1) + ", layer "+self.name)

    def _calc_eig_2d(self):
        """
        Calculate the eigen modes in purely 2D scenario, i.e. when TE and TM are separated.

        Notes
        -----
        Three conditions need to be met for this subroutine to be applicable
        *   structure is 2D, uniform in y
        *   no off-diagonal components in permittivity and permeability
        *   incident waves in x-z plane, no phi rotation.
        """
        t1 = time.process_time()

        o = self.pr.omega
        eixx = self.eixxcm
        muyy = self.muyycm
        eizz = self.eizzcm
        mixx = self.mixxcm
        epyy = self.epyycm
        mizz = self.mizzcm
        Kx = self.pr.Kx
        Ky = self.pr.Ky

        # TM is p, Ex-Hy-Ez
        Ppilu = sla.lu_factor(-1./o*eixx)  # i means inverted
        Qp = o * muyy - 1. / o * Kx[:, None] * eizz * Kx

        # TE is s, Hx-Ey-Hz
        Psilu = sla.lu_factor(-1./o * mixx)
        Qs = o * epyy - 1. / o * Kx[:, None] * mizz * Kx

        _phil = []
        _psil = []
        _ql = []
        _rc = []  # radiation channel

        for P, Q in [(Ppilu, Qp), (Psilu, Qs)]:

            ql2, phil = la.eig(- sla.lu_solve(P, Q))
            rc = np.where(ql2.real > 0)[0].tolist()  # todo: even for radiation channel, if omega.imag larger than omega.real, q02.real is negative
            ql = np.sqrt(ql2 + 0j)

            if self.in_mid_out == 'mid':
                ql[ql.imag < 0] *= -1
            else:
                if self.pr.omega.imag < 0:
                    if self.pr.ccnif == "physical":
                        ql[(ql2.real < 0) * (ql.imag < 0)] *= -1
                    elif self.pr.ccnif == "ac":
                        ql[(ql2.real < 0) * (ql.imag > 0)] *= -1
                    else:
                        warn("ccnif not recognized. Default to 'physical'.")
                        ql[(ql2.real < 0) * (ql.imag < 0)] *= -1
                elif self.pr.omega.imag > 0:
                    if self.pr.ccpif == "ac":
                        ql[(ql2.real < 0) * (ql.imag < 0)] *= -1
                    elif self.pr.ccpif == "physical":
                        ql[(ql2.real < 0) * (ql.imag > 0)] *= -1
                    else:
                        warn("ccpif not recognized. Default to 'ac'.")
                        ql[(ql2.real < 0) * (ql.imag < 0)] *= -1
                else:
                    ql[ql.imag < 0] *= -1
            ql_inv = 1. / ql
            psil = -1j * Q @ phil * ql_inv
            # psil = 1j * la.inv(P) @ phil @ np.diag(self.ql)

            _phil.append(phil)
            _psil.append(psil)
            _ql.append(ql)
            _rc.append(rc)

        self.phil = np.zeros((2*self.pr.num_g, 2*self.pr.num_g), dtype=complex)
        self.phil[:self.pr.num_g, :self.pr.num_g] = _psil[0]
        self.phil[self.pr.num_g:, self.pr.num_g:] = _phil[1]

        self.psil = np.zeros((2*self.pr.num_g, 2*self.pr.num_g), dtype=complex)
        self.psil[:self.pr.num_g, self.pr.num_g:] = _psil[1]
        self.psil[self.pr.num_g:, :self.pr.num_g] = _phil[0]

        self.ql = np.concatenate(_ql)
        self._rad_cha = _rc[0] + [a+self.pr.num_g for a in _rc[1]]

        if self.pr.show_calc_time:
            print("{:.6f}   _calc_eig_2d".format(time.process_time() - t1) + ", layer "+self.name)

    def _calc_im(self):
        """Calculate the interface matrix."""

        t1 = time.process_time()

        if self.is_vac:
            self.im = self.pr.im0
        else:
            al0, bl0 = im(self.phil, self.psil, self.pr.phi0, self.pr.psi0, self._phil_is_idt)
            self.im = (al0, bl0)

        if self.pr.show_calc_time:
            print("{:.6f}   _calc_im".format(time.process_time() - t1) + ", layer "+self.name)

    def _calc_sm(self):
        """ calculate the scattering matrix of current layer """

        t1 = time.process_time()

        if self.is_vac and self.thickness == 0:
            self.sm = self.pr.sm0
        else:
            if self.in_mid_out == 'mid':
                if self.thickness == 0:
                    sm = self.pr.sm0
                else:
                    sm = s_1l(self.thickness, self.ql, *self.im)
            elif self.in_mid_out == 'in':
                sm = s_1l_in(*self.im)
            elif self.in_mid_out == 'out':
                sm = s_1l_out(*self.im)
            else:
                raise Exception('Layer is not the incident, a middle, or the output layer.')
            self.sm = sm

        if self.pr.show_calc_time:
            print("{:.6f}   _calc_sm layer".format(time.process_time() - t1) + ", layer "+self.name)

    def solve(self):
        """
        calculate the eigen, interface matrix, and scattering matrix of the layer in vacuum
        Returns
        -------

        """
        t1 = time.process_time()

        if self.if_mod:
            self._calc_ep_mu_fs_3d()
            self._cons_ep_mu_cm_3d()
            if self.patterns:
                if self.pr.is_1d_latt and self.pr.phi == 0. and not self.materials_ode:
                    self._calc_eig_2d()
                else:
                    self._calc_PQ_3d()
                    self._calc_eig_3d()
            else:
                self._calc_eig_3d_uniform()
            self._calc_im()
        if self.if_mod or self.if_t_change:
            self._calc_sm()
            self.if_mod = False
            self.if_t_change = False

        if self.pr.show_calc_time:
            print('{:.6f}'.format(time.process_time() - t1) + '   layer ' + self.name + ' solve total')




# -*- coding: utf-8 -*-

import numpy as np


def poly_area(vertices):
    """

    Parameters
    ----------
    vertices        :   list[tuple(float, float)]
                        the vertices of the polygon ccw order

    Returns
    -------
    a               :   float
                        area of polygon
    """
    verti = np.array(vertices)
    x = verti[:, 0]
    y = verti[:, 1]

    a = 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))

    return a



# -*- coding: utf-8 -*-

import numpy as np
import numpy.linalg as la
import scipy.linalg as sla
import time
# import scipy.sparse as sps
# import warnings


def rsp(sa11, sa12, sa21, sa22, sb11, sb12, sb21, sb22):
    """
    Take the Redheffer star product (rsp) of two scattering matrices.

    Parameters
    ----------
    sa11, sa12, sa21, sa22, sb11, sb12, sb21, sb22      :   ndarray
                                                            scattering matrix A and B

    Returns
    -------
    s11     :   ndarray
    s12     :   ndarray
    s21     :   ndarray
    s22     :   ndarray

    References
    ----------

    UTEP EMLab.
    Victor's Notes on Redheffer star product is actually skewed.
    """
    time1 = time.process_time()

    # identity matrix
    # idt = np.diag(np.ones(sa11.shape[0]))
    idt = np.eye(sa11.shape[0], dtype=complex)

    # UTEP CEM (correct)
    # (bi, ao) = S (ai, bo), i, o means left 'input' region and right 'output' region, a means right going, b means left going
    # iv12 = la.inv(idt - sb11 @ sa22)
    # iv21 = la.inv(idt - sa22 @ sb11)
    #
    # s11 = sa11 + sa12 @ iv12 @ sb11 @ sa21
    # s12 = sa12 @ iv12 @ sb12
    # s21 = sb21 @ iv21 @ sa21
    # s22 = sb22 + sb21 @ iv21 @ sa22 @ sb12

    t1 = idt - sb11 @ sa22
    t2 = idt - sa22 @ sb11
    p1 = sla.solve(t1.T, sa12.T).T
    p2 = sla.solve(t2.T, sb21.T).T

    s11 = sa11 + p1 @ sb11 @ sa21
    s12 = p1 @ sb12
    s21 = p2 @ sa21
    s22 = sb22 + p2 @ sa22 @ sb12

    # Victor's, and several other online notes.
    # This is skewed, i.e. s11 is transmission, s12 is reflection.
    # i.e. (ao, bi) = S (ai, bo)
    # s21 = sa21 + sa22 @ la.inv(idt - sb21 @ sa12) @ sb21 @ sa11
    # s22 = sa22 @ la.inv(idt - sb21 @ sa12) @ sb22
    # s11 = sb11 @ la.inv(idt - sa12 @ sb21) @ sa11
    # s12 = sb12 + sb11 @ la.inv(idt - sa12 @ sb21) @ sa12 @ sb22

    print('RSP', time.process_time() - time1)

    return s11, s12, s21, s22


def rsp_in(sa11, sa12, sa21, sa22, sb11, sb12, sb21, sb22):
    """
    Redheffer star product (rsp) of two scattering matrices, assuming it's for the incident region to other layers.

    Parameters
    ----------
    sa11    :   np.ndarray
    sa12    :   np.ndarray
    sa21    :   Tuple[np.ndarray, np.ndarray]
                assuming this is the LU decomposition of Al0 for the incident layer.
    sa22    :   np.ndarray
    sb11    :   np.ndarray
    sb12    :   np.ndarray
    sb21    :   np.ndarray
    sb22    :   np.ndarray

    Returns
    -------
    s11     :   np.ndarray
    s12     :   np.ndarray
    s21     :   np.ndarray
    s22     :   np.ndarray

    """
    idt = np.eye(sa11.shape[0], dtype=complex)

    t1 = idt - sb11 @ sa22
    t2 = idt - sa22 @ sb11
    p1 = sla.solve(t1.T, sa12.T).T
    p2 = sla.solve(t2.T, sb21.T).T

    s11 = sa11 + p1 @ sla.lu_solve(sa21, sb11.T).T
    s12 = p1 @ sb12
    s21 = sla.lu_solve(sa21, p2.T).T
    s22 = sb22 + p2 @ sa22 @ sb12

    return s11, s12, s21, s22


def rsp_out(sa11, sa12, sa21, sa22, sb11, sb12, sb21, sb22):
    """
    Redheffer star product (rsp) of two scattering matrices, assuming it's for other layers to the output region.

    Parameters
    ----------
    sa11    :   np.ndarray
    sa21    :   np.ndarray
    sa12    :   np.ndarray
    sa22    :   np.ndarray
    sb11    :   np.ndarray
    sb12    :   Tuple[np.ndarray, np.ndarray]
                assuming this is the LU decomposition of Al0 for the incident layer.
    sb21    :   np.ndarray
    sb22    :   np.ndarray

    Returns
    -------
    s11     :   np.ndarray
    s12     :   np.ndarray
    s21     :   np.ndarray
    s22     :   np.ndarray

    """
    idt = np.eye(sa11.shape[0], dtype=complex)

    t1 = idt - sb11 @ sa22
    t2 = idt - sa22 @ sb11
    p1 = sla.solve(t1.T, sa12.T).T
    p2 = sla.solve(t2.T, sb21.T).T

    s11 = sa11 + p1 @ sb11 @ sa21
    s12 = sla.lu_solve(sb12, p1.T).T
    s21 = p2 @ sa21
    s22 = sb22 + p2 @ sla.lu_solve(sb12, sa22.T).T

    return s11, s12, s21, s22



# -*- coding: utf-8 -*-

import numpy as np
import numpy.linalg as la
# from scipy import sparse as sps
# import scipy.fft as fft
from typing import Tuple, List, Union, Optional, Set
# import time
from warnings import warn
from inkstone.recipro import recipro
from inkstone.g_pts import g_pts
from inkstone.g_pts_1d import g_pts_1d
from inkstone.max_idx_diff import max_idx_diff
from inkstone.conv_mtx_idx import conv_mtx_idx_2d


class Params:
    """
    Contains various parameters that are relevant to solver and shared between solver and each layer.
    """

    def __init__(self,
                 latt_vec=None,
                 num_g=None,
                 frequency=None,
                 omega=None,
                 theta=None,
                 phi=None,
                 show_calc_time = False
                 ):
        """

        Parameters
        ----------
        latt_vec        :   tuple[tuple[float, float], tuple[float, float]]
                            primitive lattice vectors
        num_g           :   int
                            number of g points
        frequency       :   float or complex
        omega           :   float or complex
        theta, phi      :   float
                            incident angles in degrees.
                            Theta is the angle between incident k and z axis.
                            phi is the angle between the in-plane projection of k and the x axis. Rotating from kx axis phi degrees ccw around z axis to arrive at the kx of incident wave.
        """

        self.gs: Optional[List[Tuple[float, float]]] = None  # list of g points for E and H fields. Added by k_inci to get the ks, i.e. k points for E and H
        self.idx_g: Optional[List[Tuple[int, int]]] = None  # list of g points indices
        self.idx_conv_mtx: Optional[np.ndarray] = None  # indexing array to for constructing convolution matrices.
        self.ks: Optional[List[Tuple[float, float]]] = None  # list of k points for E and H fields.
        self.idx_g_ep_mu: Optional[List[Tuple[int, int]]] = None
        self.idxa_g_ep_mu: Optional[Tuple[np.ndarray, np.ndarray]] = None  # The g indices for ep and mu.
        self.idx_g_ep_mu_used: Optional[List[Tuple[int, int]]] = None  # the actually used epsilon and mu grid points
        self.ks_ep_mu: Optional[List[Tuple[float, float]]] = None  # list of k points where epsi and mu Fourier components are to be calculated
        self.ka_ep_mu: Optional[Tuple[np.ndarray, np.ndarray]] = None  # The k points for ep and mu, Tuple of meshgrid (kx, ky).
        self.Kx: Optional[np.ndarray] = None  # the diagonals of the Big diagonal matrix
        self.Ky: Optional[np.ndarray] = None  # the diagonals of the Big diagonal matrix
        self.Kz: Optional[np.ndarray] = None  # the diagonals of the Big diagonal matrix, useful for uniform layers
        self.mmax: Optional[int] = None  # max index in b1 direction for epsi and mu Fourier series
        self.nmax: Optional[int] = None  # max index in b2 direction for epsi and mu Fourier series
        self.phi0: Optional[np.ndarray] = None  # E field eigen mode in vacuum (identity matrix), side length 2*num_g
        self.psi0: Optional[np.ndarray] = None  # H field eigen mode in vacuum (Q * Phi * q^-1), side length 2*num_g
        self.q0: Optional[np.ndarray] = None  # 1d array, eigen propagation constant in z direction in vacuum, length 2*num_g
        self.P0_val: Optional[Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]] = None  # Tuple of 4, each is an ndarray of size num_g, containing the diagonal elements of the 4 blocks of P0.
        self.Q0_val: Optional[Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]] = None  # Tuple of 4, each is an ndarray of size num_g, containing the diagonal elements of the 4 blocks of Q0.
        self.im0: Optional[Tuple[np.ndarray, np.ndarray]] = None  # vacuum interface matrix
        self.sm0: Optional[Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]] = None  # vacuum scattering matrix, (s11_0, s12_0, s21_0, s22_0), each of s_ij has side length 2*num_g
        self._rad_cha_0: Optional[List[int]] = None

        self.ccnif = "physical"
        self.ccpif = "ac"

        self.__num_g_actual: Optional[int] = None  # actual number of G points used.
        self._num_g_input: Optional[int] = None  # number of G points input by user
        self._k_inci: Optional[Tuple[Tuple[float, float]], Tuple[float, float]] = None  # incident in-plane k
        self._omega: Optional[complex] = None
        self._frequency: Optional[complex] = None
        self._recipr_vec: Optional[Tuple[Tuple[float, float], Tuple[float, float]]] = None
        self._latt_vec: Optional[Tuple[Tuple[float, float], Tuple[float, float]]] = None
        self.is_1d_latt: bool = False  # if this is solving a 2d structure. Incident wave can still have phi rotation, i.e. 3D scenario for a 2D structure.
        self._theta: Optional[float] = None  # in rad
        self._phi: Optional[float] = None  # in rad
        self._uc_area: Optional[float] = None
        self.cos_varthetas: Optional[List[float]] = None  # incident angles
        self.sin_varthetas: Optional[List[float, complex]] = None
        self.cos_phis: Optional[List[float]] = None
        self.sin_phis: Optional[List[float]] = None

        self._incident_orders = None
        self._incident_orders_bk = None
        self._s_amps = None
        self._p_amps = None
        self._s_amps_bk = None
        self._p_amps_bk = None
        self.ai: Optional[np.ndarray] = None
        self.bo: Optional[np.ndarray] = None

        self.q0_contain_0: bool = False

        self.show_calc_time = show_calc_time

        # property initialization can change other attributes hence must be after initialization of other attributes.
        if latt_vec is not None:
            self.latt_vec = latt_vec
        if num_g is not None:
            self.num_g = num_g
        if frequency is not None:
            self.frequency = frequency
        if omega is not None:
            self.omega = omega
        if theta is not None:
            self.theta = theta
        if phi is not None:
            self.phi = phi

    @property
    def latt_vec(self) -> Union[float, Tuple[Tuple[float, float], Tuple[float, float]]]:
        """
        Primitive lattice vectors.
        The lattice vector and be a single number which means 1D, or tuples such as ((a1x, a1y), (a2x, a2y)) which means 2D .

        If tuples are supplied but one of them is (0, 0), then it is also a 1d. In this case the lattice will be rotated such that the actual 1d direction is in x axis where y is uniform.
        Note the incident wave angles will be w.r.t. the structure AFTER it's rotated to x axis.

        """
        return self._latt_vec

    @latt_vec.setter
    def latt_vec(self, val: Union[float, Tuple[Tuple[float, float], Tuple[float, float]]]):
        if type(val) is not tuple:
            val = ((val, 0), (0, 0))
            self.is_1d_latt = True
        else:
            a, b = val
            an, bn = [la.norm(a), la.norm(b)]
            # if either latt vec is zero, then it is 1D. Make sure it's x-z for layer solver.
            if an == 0:
                val = ((bn, 0), (0, 0))
                self.is_1d_latt = True
                warn("2D structure (in-plane 1D), non-uniform direction is rotated to x axis.")
            if bn == 0:
                val_old = val
                val = ((an, 0), (0, 0))
                self.is_1d_latt = True
                if val_old != val:
                    warn("2D structure (in-plane 1D), non-uniform direction is rotated to x axis.")
        self._latt_vec: Union[Tuple[Tuple[float, float], Tuple[float, float]]] = val
        self._recipr_vec: Tuple[Tuple[float, float], Tuple[float, float]] = recipro(val[0], val[1])
        self.if_2d()
        self._calc_gs()
        self._calc_uc_area()

    @property
    def recipr_vec(self) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        """Reciprocal lattice vectors"""
        return self._recipr_vec

    def if_2d(self):
        b1, b2 = self._recipr_vec
        if b1[0] == float('inf') or b1[1] == float('inf') or b2[0] == float('inf') or b2[1] == float('inf'):
            self.is_1d_latt = True
        else:
            self.is_1d_latt = False

    @property
    def num_g(self) -> int:
        """The number of G points to be included. To ensure symmetry, the actual num_g may be slightly different than the user set value."""
        return self._num_g_ac

    @num_g.setter
    def num_g(self, val: int):
        self._num_g_input: int = val
        self._calc_gs()

    @property
    def _num_g_ac(self) -> int:
        """actual number of G points"""
        return self.__num_g_actual

    @_num_g_ac.setter
    def _num_g_ac(self, val: int):
        # todo: all other calc should use this than original num_g
        self.__num_g_actual = val
        self._calc_phi0()
        # self._calc_q0()  # called through _calc_gs - _calc_ks
        self._calc_im0()
        self._calc_s_0()
        # self.calc_ai_bo_3d()  # called through _calc_gs

    @property
    def frequency(self) -> Union[float, complex]:
        """ frequency, c/lambda in vacuum """
        return self._frequency

    @frequency.setter
    def frequency(self, val: Union[float, complex]):
        if val is not None:
            self.omega = val * np.pi * 2.
            self._frequency = val

    @property
    def omega(self) -> Union[float, complex]:
        """ angular frequency, 2pi c/lambda in vacuum """
        return self._omega

    @omega.setter
    def omega(self, val: Union[float, complex]):
        if val is not None:
            self._omega = val
            self._frequency = val / np.pi / 2.
            self.q0_contain_0 = False
            self._calc_k_inci()
            # self._calc_q0()  # called through _calc_k_inci - _calc_ks
            # self._calc_P0Q0()  # called through _calc_k_inci - _calc_ks - _calc_Km
            # self._calc_angles()  # called through _calc_k_inci - _calc_ks -

    @property
    def k_inci(self) -> Tuple[float, float]:
        """kx and ky project of the incident wave vector"""
        return self._k_inci

    @property
    def theta(self) -> Union[float, complex]:
        """he angle between incident k and z axis, in degrees, range: [0, pi/2]"""
        if self._theta is not None:
            theta = self._theta / np.pi * 180.
        else:
            theta = None
        return theta

    @theta.setter
    def theta(self, val: Union[float, complex]):
        if val is not None:
            self._theta: float = val * np.pi / 180.
            self._calc_k_inci()
            self._calc_angles()
        else:
            if self._theta is None:
                self._theta = 0.

    @property
    def phi(self) -> float:
        """ the angle between the in-plane projection of k and the x axis in degrees. Rotating from kx axis phi degrees ccw around z axis to arrive at the kx of incident wave. """
        if self._phi is not None:
            phi = self._phi / np.pi * 180.
        else:
            phi = None
        return phi

    @phi.setter
    def phi(self, val: float):
        if val is not None:
            self._phi: float = val * np.pi / 180.
            self._calc_k_inci()
            self._calc_angles()
        else:
            if self._phi is None:
                self._phi = 0.

    @property
    def rad_cha_0(self) -> List[int]:
        """list of indices in `self.idx_g` that corresponds to radiation channels in vacuum"""
        return self._rad_cha_0  # between 0 and num_g

    @property
    def ccnif(self) -> str:
        """
        {"physical", "ac"}, default: "physical".
        """
        return self._ccnif

    @ccnif.setter
    def ccnif(self, val):
        if (val == "physical") or (val == "ac"):
            self._ccnif = val
        else:
            warn("ccnif not understood and not changed in the solver.")

    @property
    def ccpif(self) -> str:
        """
        {"physical", "ac"}, default: "ac".
        """
        return self._ccpif

    @ccpif.setter
    def ccpif(self, val):
        if (val == "physical") or (val == "ac"):
            self._ccpif = val
        else:
            warn("ccpif not understood and not changed in the solver.")

    def set_inci_ord_amp(self,
                         s_amplitude: Optional[Union[float, complex, List[Union[float, complex]]]] = None,
                         p_amplitude: Optional[Union[float, complex, List[Union[float, complex]]]] = None,
                         order: Union[Tuple[int, int], List[Tuple[int, int]]] = None,
                         s_amplitude_back: Union[float, List[float]] = None,
                         p_amplitude_back: Union[float, List[float]] = None,
                         order_back: Union[Tuple[int, int], List[Tuple[int, int]]] = None
                         ):
        """
        Setting and resetting incident orders and amplitudes.

        If any of the optional arguments are not given, then use the previously set values.

        Parameters
        ----------
        s_amplitude
        p_amplitude
        order
        s_amplitude_back
        p_amplitude_back
        order_back
        """
        # Convert order and order_back to lists (can be empty list)
        # todo: how do I know the user chosen order is in idx_g?
        if order is None:
            if self._incident_orders is None:
                self._incident_orders = [(0, 0)]
        elif type(order) is tuple:
            order = [order]
            self._incident_orders = order
        else:
            self._incident_orders = order

        if order_back is None:
            if self._incident_orders_bk is None:
                self._incident_orders_bk = [(0, 0)]
        elif type(order_back) is tuple:
            order_back = [order_back]
            self._incident_orders_bk = order_back
        else:
            self._incident_orders_bk = order_back

        # convert s and p amplitudes to lists (can be empty list)
        amp = []
        for amps, selfamps, od in zip([[s_amplitude, p_amplitude], [s_amplitude_back, p_amplitude_back]], [[self._s_amps, self._p_amps], [self._s_amps_bk, self._p_amps_bk]], [self._incident_orders, self._incident_orders_bk]):
            for a, s in zip(amps, selfamps):
                if a is None:
                    if s is None:
                        a = [0. for i in range(len(od))]
                    else:
                        a = s
                elif not hasattr(a, "__len__"):
                    a = [a]
                amp.append(a)
        self._s_amps = amp[0]
        self._p_amps = amp[1]
        self._s_amps_bk = amp[2]
        self._p_amps_bk = amp[3]

        if (len(self._incident_orders) != len(self._s_amps)) or (len(self._incident_orders) != len(self._p_amps)):
            raise Exception("The list length of the incident s amplitudes, p amplitudes, and incident orders are not equal.")

        if (len(self._incident_orders_bk) != len(self._s_amps_bk)) or (len(self._incident_orders_bk) != len(self._p_amps_bk)):
            raise Exception("The list length of the backside incident s amplitudes, p amplitudes, and incident orders are not equal.")

        self.calc_ai_bo_3d()

    def _calc_k_inci(self):
        if (self.theta is not None) and (self.phi is not None) and (self.omega is not None):
            kx = self.omega.real * np.cos(np.pi/2 - self._theta) * np.cos(self._phi)
            ky = self.omega.real * np.cos(np.pi/2 - self._theta) * np.sin(self._phi)
            # todo: using self.omega (kx ky complex) also gives answers, physical meaning is different
            self._k_inci: Tuple[float, float] = (kx, ky)
            self._calc_ks()

    def _calc_gs(self):
        """ calculate E and H Fourier components g points """
        if self._num_g_input and self.recipr_vec:
            b1, b2 = self.recipr_vec
            b1n, b2n = [la.norm(b) for b in [b1, b2]]
            if b1n != float('inf') and b2n != float('inf'):
                self.gs, self.idx_g = g_pts(self._num_g_input, self.recipr_vec[0], self.recipr_vec[1])
            elif b1n == float('inf'):
                self.gs, idx = g_pts_1d(self._num_g_input, b2)
                self.idx_g = [(0, i) for i in idx]
            elif b2n == float('inf'):
                self.gs, idx = g_pts_1d(self._num_g_input, b1)
                self.idx_g = [(i, 0) for i in idx]
            else:
                raise Exception("Both reciprocal lattice vectors are infinite. Can't calculate g points.")
            self._num_g_ac = len(self.gs)
            self._calc_ks()
            self._calc_conv_mtx_idx()
            # self.calc_ai_bo_3d()  # called through _calc_ks() - _calc_angles()

    def _calc_ks(self):
        if self.gs and self.k_inci:
            self.ks = [(g[0]+self.k_inci[0], g[1] + self.k_inci[1]) for g in self.gs]
            self._calc_Km()
            self._calc_ks_ep_mu()
            self._calc_q0()
            self._calc_angles()

    def _calc_Km(self):
        """Calculate Kx Ky matrices"""
        if self.ks:
            # t1 = time.process_time()
            ksa = np.array(self.ks)  # Nx2 shape
            kx = ksa[:, 0]
            ky = ksa[:, 1]
            self.Kx = kx
            self.Ky = ky

            # print('_calc_Km', time.process_time() - t1)

            self._calc_P0Q0()

    def _calc_conv_mtx_idx(self):
        if self.idx_g:
            self.idx_conv_mtx = conv_mtx_idx_2d(self.idx_g, self.idx_g)
            self.idx_g_ep_mu_used = list(set([(i, j) for (i, j) in self.idx_conv_mtx.reshape(self._num_g_ac ** 2, 2)]))  # The first element of each tuple is in physical "x" direction

    def _calc_ks_ep_mu(self):
        if self.idx_g:
            # t1 = time.process_time()

            m, n = max_idx_diff(self.idx_g)
            self.mmax = m
            self.nmax = n
            x = np.arange(-m, m+1)
            y = np.arange(-n, n+1)
            xx, yy = np.meshgrid(x, y)

            # t2 = time.process_time()

            self.idx_g_ep_mu = list(zip(xx.ravel(), yy.ravel()))
            self.idxa_g_ep_mu = (xx, yy)

            # t3 = time.process_time()

            b1, b2 = self.recipr_vec
            b1n, b2n = [la.norm(b) for b in [b1, b2]]
            # if one of them is infinity, this is a 1D structure. set it to 0 such that they don't appear in ks_ep_mu
            if b1n == float('inf'):
                b1 = (0, 0)
            if b2n == float('inf'):
                b2 = (0, 0)

            kx = xx * b1[0] + yy * b2[0]
            ky = xx * b1[1] + yy * b2[1]
            self.ks_ep_mu = list(zip(kx.ravel(), ky.ravel()))  # each element is (kx, ky) tuple.
            self.ka_ep_mu = (kx, ky)

            # print(t2 - t1)
            # print(t3 - t2)
            # print(time.process_time() - t3)
            # print('_calc_ks_ep_mu', time.process_time()-t1)

    def _calc_q0(self):
        if self._num_g_ac and (self.omega is not None) and self.ks:
            # t1 = time.process_time()
            k_parallel = np.linalg.norm(np.array(self.ks), axis=-1)
            q02 = np.ones(self._num_g_ac) * np.square(self.omega) - np.square(k_parallel) + 0j
            self._rad_cha_0 = np.where(q02.real > 0)[0].tolist()  # todo: even for radiation channel, if omega.imag larger than omega.real, q02.real is negative
            q0 = np.sqrt(q02)
            if self.omega.imag < 0:
                if self.ccnif == "physical":
                    q0[(q02.real < 0) * (q0.imag < 0)] *= -1
                elif self.ccnif == "ac":
                    q0[(q02.real < 0) * (q0.imag > 0)] *= -1
                else:
                    warn("ccnif not recognized. Default to 'physical'.")
                    q0[(q02.real < 0) * (q0.imag < 0)] *= -1
            elif self.omega.imag > 0:
                if self.ccpif == "ac":
                    q0[(q02.real < 0) * (q0.imag < 0)] *= -1
                elif self.ccpif == "physical":
                    q0[(q02.real < 0) * (q0.imag > 0)] *= -1
                else:
                    warn("ccpif not recognized. Default to 'ac'.")
                    q0[(q02.real < 0) * (q0.imag < 0)] *= -1
            else:
                q0[q0.imag < 0] *= -1
            # todo: what to do at q02.real == 0 case (Woods)
            self.q0 = np.concatenate([q0, q0])
            # print('_calc_q0', time.process_time() - t1)
            self._calc_psi0()

    def _calc_P0Q0(self):
        if (self.omega is not None) and (self.Kx is not None) and (self.Ky is not None):
            # t1 = time.process_time()

            Kx = self.Kx
            Ky = self.Ky
            o = self.omega

            P11 = 1. / o * Kx * Ky
            P12 = o - 1. / o * Kx * Kx
            P21 = -o + 1. / o * Ky * Ky
            P22 = - 1. / o * Ky * Kx

            # t2 = time.process_time()
            # print('P0 Q0 blocks', t2 - t1)

            self.P0_val = (P11, P12, P21, P22)
            self.Q0_val = self.P0_val

            self._calc_psi0()

    def _calc_phi0(self):
        if self._num_g_ac:
            self.phi0 = np.eye(2 * self._num_g_ac, 2 * self._num_g_ac, dtype=complex)

    def _calc_psi0(self):
        if (self.Q0_val is not None) and (self.q0 is not None) and (self._num_g_ac is not None):
            # t1 = time.process_time()
            with np.errstate(divide='raise'):
                try:
                    q0_inv = 1. / self.q0

                    ng = self._num_g_ac
                    psi0 = np.zeros((2 * ng, 2 * ng), dtype=complex)
                    r1 = range(ng)
                    r2 = range(ng, 2 * ng)
                    if self.Q0_val[0].size == q0_inv[:ng].size:
                        psi0[r1, r1] = -1j * self.Q0_val[0] * q0_inv[:ng]
                        psi0[r1, r2] = -1j * self.Q0_val[1] * q0_inv[ng:]
                        psi0[r2, r1] = -1j * self.Q0_val[2] * q0_inv[:ng]
                        psi0[r2, r2] = -1j * self.Q0_val[3] * q0_inv[ng:]
                    self.psi0 = psi0

                except FloatingPointError:
                    warn("Vacuum propagation constant 0 encountered. Possibly Wood's anomaly.", RuntimeWarning)
                    # print(self.frequency, self.k_inci, self.theta, self.q0, self.ks, self.idx_g)
                    self.q0_contain_0 = True

            # print('_calc_psi0', time.process_time() - t1)

    def _calc_im0(self):
        if self._num_g_ac:
            a0 = np.diag(2 * np.ones(2 * self._num_g_ac))
            b0 = np.zeros((2 * self._num_g_ac, 2 * self._num_g_ac))
            self.im0 = (a0, b0)

    def _calc_s_0(self):
        """calculate vacuum scattering matrix"""
        # todo: change into sparse to save memory
        if self._num_g_ac:
            # t1 = time.process_time()
            s11_0 = np.zeros((2 * self._num_g_ac, 2 * self._num_g_ac), dtype=complex)
            s12_0 = np.eye(2 * self._num_g_ac, dtype=complex)
            s21_0 = s12_0
            s22_0 = s11_0
            self.sm0 = (s11_0, s12_0, s21_0, s22_0)
            # print('_calc_s_0_3d', time.process_time() - t1)

    @property
    def uc_area(self) -> float:
        """Unit cell area"""
        return self._uc_area

    def _calc_uc_area(self):
        if self.latt_vec:
            if type(self.latt_vec) is not tuple:
                self._uc_area = self.latt_vec
            else:
                a1, a2 = self.latt_vec
                a1n, a2n = [la.norm(a) for a in [a1, a2]]
                if a1n != 0. and a2n != 0.:
                    self._uc_area = np.abs(np.cross(self.latt_vec[0], self.latt_vec[1]))
                elif a1n == 0.:  # 1D
                    self._uc_area = a2n
                elif a2n == 0.:
                    self._uc_area = a1n
                else:
                    raise Exception("Both lattice vectors have zero norms. Can't calculate unit cell area.")

    def _calc_angles(self):
        """
        calculate cos(vartheta), sin(vartheta), cos(phi), sin(phi) for all relevant orders, where vartheta = pi/2-theta
        This is used for setting incidence in high orders.
        """
        if self.ks:
            # t1 = time.process_time()

            cphi = np.zeros(self._num_g_ac, complex)
            sphi = np.zeros(self._num_g_ac, complex)
            cthe = np.zeros(self._num_g_ac, complex)
            sthe = np.zeros(self._num_g_ac, complex)

            idxa = np.array(self.idx_g)
            ksa = np.array(self.ks)
            k_pa = la.norm(ksa, axis=-1)  # norm of in-plane momentum

            i0 = (k_pa == 0.)
            ib = (k_pa != 0.)

            cphi[i0] = 1.
            sphi[i0] = 0.
            cphi[ib] = ksa[ib, 0] / k_pa[ib]
            sphi[ib] = ksa[ib, 1] / k_pa[ib]
            cthe = k_pa / self.omega + 0j
            sthe = np.sqrt(1 - cthe**2 + 0j)

            # these ensure the incident channel has real cos, sin
            ii = (idxa[:, 0] == 0) & (idxa[:, 1] == 0)
            cphi[ii] = np.cos(self._phi)
            sphi[ii] = np.sin(self._phi)
            cthe[ii] = np.cos(np.pi/2 - self._theta)
            sthe[ii] = np.sin(np.pi/2 - self._theta)

            self.cos_phis = list(cphi)
            self.sin_phis = list(sphi)
            self.cos_varthetas = list(cthe)
            self.sin_varthetas = list(sthe)

            # print('_calc_angles', time.process_time()-t1)
        self.calc_ai_bo_3d()

    def calc_ai_bo_3d(self):
        """calculate incident ai and bo amplitudes"""

        # t1 = time.process_time()

        aibo = [self.ai, self.bo]
        for ii, (sa, pa, od) in enumerate([[self._s_amps, self._p_amps, self._incident_orders], [self._s_amps_bk, self._p_amps_bk, self._incident_orders_bk]]):
            if self._num_g_ac:
                ab = np.zeros(2 * self._num_g_ac) + 0j
                if (sa or pa) and od and self.idx_g and \
                        self.sin_phis and self.sin_varthetas and self.cos_phis and self.cos_varthetas:
                    # find the index of the input orders in the g list
                    idx = [i for order in od for i, j in enumerate(self.idx_g) if j == order]
                    for i in range(len(idx)):
                        s = sa[i]
                        p = pa[i]
                        ab[i] = -s * self.sin_phis[i] + p * self.sin_varthetas[i] * self.cos_phis[i]  # e_x
                        ab[i + self._num_g_ac] = s * self.cos_phis[i] + p * self.sin_varthetas[i] * self.sin_phis[i]  # e_y
                aibo[ii] = ab
        self.ai, self.bo = aibo
        # print('calc_ai_bo_3d', time.process_time() - t1)




# -*- coding: utf-8 -*-

from typing import Tuple, Optional, List
import time
import numpy as np
from inkstone.layer import Layer
from inkstone.sm import s_1l, s_1l_in, s_1l_out


class LayerCopy:

    def __init__(self, name: str, layer: Layer, thickness: float):
        """
        A dummy layer for layer copies

        Parameters
        ----------
        name        :   the name of this layer
        layer       :   the original layer
        thickness   :   thickness of this layer
        """

        self.name = name
        self.layer = layer
        self._al_bl: Optional[Tuple[np.ndarray, np.ndarray]] = None
        self.in_mid_out: str = 'mid'  # {'in', 'mid', 'out'}, if this layer is the incident, output, or a middle layer

        self.sm: Optional[Tuple[np.ndarray, np.ndarray]] = None
        self._if_t_change = True
        self.need_recalc_al_bl = True

        self.thickness = thickness

    def __getattr__(self, item):
        """ignore all calls if not explicitly defined"""
        # raise Exception('what the hell')
        # print(1+1)
        pass

    @property
    def thickness(self):
        return self._thickness

    @thickness.setter
    def thickness(self, val):
        if self._thickness != val:
            self._thickness = val
            self._if_t_change = True

    @property
    def materials_used(self):
        return self.layer.materials_used

    @property
    def if_mod(self):
        return self.layer.if_mod

    @if_mod.setter
    def if_mod(self, val):
        pass

    @property
    def im(self):
        return self.layer.im

    @property
    def ql(self):
        return self.layer.ql

    @property
    def phil(self):
        return self.layer.phil

    @property
    def psil(self):
        return self.layer.psil

    @property
    def eizzcm(self):
        return self.layer.eizzcm

    @property
    def mizzcm(self):
        return self.layer.mizzcm

    @property
    def al_bl(self) -> Optional[Tuple[np.ndarray, np.ndarray]] :
        return self._al_bl

    @al_bl.setter
    def al_bl(self, val: Tuple[np.ndarray, np.ndarray]):
        self._al_bl = val

    @property
    def is_vac(self) -> bool:
        return self.layer.is_vac

    @property
    def rad_cha(self) -> List[int]:
        return self.layer.rad_cha

    def solve(self):
        t1 = time.process_time()
        if self.layer.if_mod or self._if_t_change:
            self.layer.solve()
            self._calc_sm()
            self._if_t_change = False
        if self.layer.pr.show_calc_time:
            print('{:.6f}'.format(time.process_time() - t1) + '   layer ' + self.name+' solve (layer copy)')

    def _calc_sm(self):
        """ calculate the scattering matrix of current layer """

        t1 = time.process_time()

        if self.is_vac and self.thickness == 0:
            self.sm = self.layer.pr.sm0
        else:
            if self.in_mid_out == 'mid':
                if self.thickness == 0:
                    sm = self.layer.pr.sm0
                else:
                    sm = s_1l(self.thickness, self.layer.ql, *self.layer.im)
            elif self.in_mid_out == 'in':
                sm = s_1l_in(*self.layer.im)
            elif self.in_mid_out == 'out':
                sm = s_1l_out(*self.layer.im)
            else:
                raise Exception('Layer is not the incident, a middle, or the output layer.')
            self.sm = sm

        if self.layer.pr.show_calc_time:
            print('{:.6f}   _calc_sm  (layer copy)'.format(time.process_time() - t1))

